"""
Plot posteriors of the IMR consistency test 
"""

import numpy as np
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import scipy
import scipy.ndimage.filters as filter
import time
import os
import os.path
import commands
import pickle, gzip
from scipy import interpolate
import optparse as op
from matplotlib import rc
import matplotlib

# Module for confidence calculations
class confidence(object):
  def __init__(self, counts):
    # Sort in descending order in frequency
    self.counts_sorted = np.sort(counts.flatten())[::-1]
    # Get a normalized cumulative distribution from the mode
    self.norm_cumsum_counts_sorted = np.cumsum(self.counts_sorted) / np.sum(counts)
    # Set interpolations between heights, bins and levels
    self._set_interp()
  def _set_interp(self):
    self._length = len(self.counts_sorted)
    # height from index
    self._height_from_idx = interpolate.interp1d(np.arange(self._length), self.counts_sorted, bounds_error=False, fill_value=0.)
    # index from height
    self._idx_from_height = interpolate.interp1d(self.counts_sorted[::-1], np.arange(self._length)[::-1], bounds_error=False, fill_value=self._length)
    # level from index
    self._level_from_idx = interpolate.interp1d(np.arange(self._length), self.norm_cumsum_counts_sorted, bounds_error=False, fill_value=1.)
    # index from level
    self._idx_from_level = interpolate.interp1d(self.norm_cumsum_counts_sorted, np.arange(self._length), bounds_error=False, fill_value=self._length)
  def level_from_height(self, height):
    return self._level_from_idx(self._idx_from_height(height))
  def height_from_level(self, level):
    return self._height_from_idx(self._idx_from_level(level))

# gaussian filter with 1-sigma softening
def gf(P):
	return filter.gaussian_filter(P, sigma=1.0)

# compute 1-sigma confidence intervals in 1D
def calc_cred_intervals_in_1d(P, x):

	# find the value of P corresponding to 50% and 9% confidence heights
	conf = confidence(P)
	P_s1 = conf.height_from_level(0.5)
	P_s2 = conf.height_from_level(0.9)

	# calculation of condifence edges (values of x corresponding to the height s1 on the two sides)
	x_s1_l = min(x[np.where(P >= P_s1)[0]])
	x_s1_r = max(x[np.where(P >= P_s1)[0]])

	# calculation of condifence edges (values of x corresponding to the height s2 on the two sides)
	x_s2_l = min(x[np.where(P >= P_s2)[0]])
	x_s2_r = max(x[np.where(P >= P_s2)[0]])

	return P_s1, P_s2, x_s1_l, x_s1_r, x_s2_l, x_s2_r

# main program
if __name__ == "__main__":

	# Choose a waveform approximant 
	approx = 'SEOBNRv4'
	approx = 'IMRPhenomPv2'

	event_data = {'GW150914': {'color':'#00a7f0', 'lw':2, 'ls':[2,0.1]}, 'GW170104': {'color':'#55b300','lw':2, 'ls':[2,0.5]}, 'GW170729': {'color':'#ea65ff','lw':1.5, 'ls':[2,0.1]}, 'GW170809': {'color':'#00b1a4','lw':1.5, 'ls':[8,0.5]}, 'GW170814': {'color':'#58ae87','lw':1, 'ls':[2,0.1]}, 'GW170818': {'color':'#5ca9b8','lw':1, 'ls':[4,1]}, 'GW170823': {'color':'#9e94e2','lw':1, 'ls':[2,1]}}

	# number of bins to evaluate the posterior distributions 
	N_bins = 401 

	# initialize variables
	P_dMfbyMf_dchifbychif_joint = np.ones((N_bins, N_bins))

	# start figure 
	p = plt.figure(figsize=(5,5))
	ax1 = plt.subplot2grid((3,3), (0,0), colspan=2)
	ax2 = plt.subplot2grid((3,3), (1,2), rowspan=2)
	ax3 = plt.subplot2grid((3,3), (1,0), colspan=2, rowspan=2)

	# loop over different events 
	for event in np.sort(event_data.keys()):

		# read the posterior data for this event
		P_dMfbyMf_dchifbychif = np.loadtxt('P_dMfbyMf_dchifbychif_%s_%s.dat.gz' %(event, approx))
		dMfbyMf_vec = np.loadtxt('dMfbyMf_vec_%s_%s.dat.gz' %(event, approx))
		dchifbychif_vec = np.loadtxt('dchifbychif_vec_%s_%s.dat.gz' %(event, approx))

		P_dMfbyMf_dchifbychif_interp_obj = scipy.interpolate.interp2d(dMfbyMf_vec, dchifbychif_vec, P_dMfbyMf_dchifbychif, fill_value=0., bounds_error=False)	

		# interpolate to common grid 
		dMfbyMf_vec = np.linspace(-1.5, 1.5, N_bins)
		dchifbychif_vec = np.linspace(-1.0, 1.0, N_bins)
		P_dMfbyMf_dchifbychif = P_dMfbyMf_dchifbychif_interp_obj(dMfbyMf_vec, dchifbychif_vec)

		# normalize the 2D posterior 
		dx = np.mean(np.diff(dMfbyMf_vec))
		dy = np.mean(np.diff(dchifbychif_vec))
		P_dMfbyMf_dchifbychif /= np.sum(P_dMfbyMf_dchifbychif) * dx * dy

		# compute the credible regions of this posterior
		conf_v1v2 = confidence(P_dMfbyMf_dchifbychif)
		s1_v1v2 = conf_v1v2.height_from_level(0.5)
		s2_v1v2 = conf_v1v2.height_from_level(0.9)

		# compute marginalized 1D posterior 
		P_dMfbyMf = np.sum(P_dMfbyMf_dchifbychif, axis=0) * dy
		P_dchifbychif = np.sum(P_dMfbyMf_dchifbychif, axis=1) * dx
		P_dMfbyMf /= np.sum(P_dMfbyMf) * dx
		P_dchifbychif /= np.sum(P_dchifbychif) * dy

		# calculate the credible levels and intervals in the marginalized 1d posteriors (this event)
		s1_v1, s2_v1, left1_v1, right1_v1, left2_v1, right2_v1 = calc_cred_intervals_in_1d(P_dMfbyMf, dMfbyMf_vec)
		s1_v2, s2_v2, left1_v2, right1_v2, left2_v2, right2_v2 = calc_cred_intervals_in_1d(P_dchifbychif, dchifbychif_vec)

		gr_height = P_dMfbyMf_dchifbychif[np.argmin(abs(dMfbyMf_vec)), np.argmin(abs(dchifbychif_vec))]
		gr_credib_level = conf_v1v2.level_from_height(gr_height)
		print 'Credible level of the GR value (%s): %.1f%% '%(event, 100.*gr_credib_level)
		print '90%% credible intervals on dMfbyMfmean (%s): %2.3f, %2.3f '%(event, left2_v1, right2_v1)
		print '90%% credible intervals on dchifbychifmean (%s): %2.3f, %2.3f '%(event, left2_v2, right2_v2)

		# plot the posteriors of this event 
		ax1.plot(dMfbyMf_vec, P_dMfbyMf, color=event_data[event]['color'], label=event, lw=event_data[event]['lw'], dashes=event_data[event]['ls'])
		ax2.plot(P_dchifbychif, dchifbychif_vec, color=event_data[event]['color'], label=event, lw=event_data[event]['lw'], dashes=event_data[event]['ls'])
		CS = ax3.contour(dMfbyMf_vec,dchifbychif_vec, gf(P_dMfbyMf_dchifbychif), levels=(s2_v1v2,), linewidths=(event_data[event]['lw'],), colors=event_data[event]['color'], linestyles='dashed')
		for c in CS.collections:
			c.set_dashes([(0, event_data[event]['ls'])])

		# load the pickle file containing the interpolant of the prior in delta_Mf/Mf, delta_af/a
		prior_file = 'Prior_dMfbyMf_dchifbychif_%s_%s' %(event, approx)
		f = gzip.open(prior_file+".pklz",'rb')
		P_dMfbyMf_dchifbychif_pr_interp_obj = pickle.load(f)

		# evaluate the prior interpolant
		P_dMfbyMf_dchifbychif_pr = P_dMfbyMf_dchifbychif_pr_interp_obj(dMfbyMf_vec, dchifbychif_vec)
		P_dMfbyMf_dchifbychif_pr /= np.sum(P_dMfbyMf_dchifbychif_pr) * dx * dy

		# removing nans
		P_dMfbyMf_dchifbychif[np.isnan(P_dMfbyMf_dchifbychif)] = 0.
		P_dMfbyMf_dchifbychif[np.isinf(P_dMfbyMf_dchifbychif)] = 0.

		# replace all zeros in the posterior vector by a small number.
		# This is to avoid multiplication by zeros while combining the posteiror
		zidx = np.where(P_dMfbyMf_dchifbychif == 0.)
		P_dMfbyMf_dchifbychif[zidx] = 1e-16

		# only use events that cross the IFAR > 1000 threshold for computing combined posteriors 
		if (event == "GW170729" or event == "GW170818"):
			print "%s: not used for joint posterior computation"%event
		else:
			print "%s: used for joint posterior computation"%event

			# Joint Probability distribution computation. From the second event onwards, divide the posterior by the corresponding
			# prior so that we are multiplying the previous posterior with the likelihood of the current event
			if event == 'GW170823':
				P_dMfbyMf_dchifbychif_joint *= P_dMfbyMf_dchifbychif
				P_dMfbyMf_dchifbychif_pr_plot = P_dMfbyMf_dchifbychif_pr
			else:
				P_dMfbyMf_dchifbychif_joint *= P_dMfbyMf_dchifbychif/P_dMfbyMf_dchifbychif_pr

			# removing nans and inf, normalizing the joint posterior 
			P_dMfbyMf_dchifbychif_joint[np.isnan(P_dMfbyMf_dchifbychif_joint)] = 0.
			P_dMfbyMf_dchifbychif_joint[np.isinf(P_dMfbyMf_dchifbychif_joint)] = 0.
			P_dMfbyMf_dchifbychif_joint /= np.sum(P_dMfbyMf_dchifbychif_joint) * dx * dy

			# END OF LOOP 

	################################################## 
	# compute final results from combined posteriors # 
	##################################################
 
	# Marginalization to one-dimensional prior distributions 
	P_dMfbyMf_pr = np.sum(P_dMfbyMf_dchifbychif_pr_plot, axis=0) * dy
	P_dchifbychif_pr = np.sum(P_dMfbyMf_dchifbychif_pr_plot, axis=1) * dx

	# normalisation of prior distributions 
	P_dMfbyMf_pr /= np.sum(P_dMfbyMf_pr) * dx
	P_dchifbychif_pr /= np.sum(P_dchifbychif_pr) * dy

	# Calculation of credible levels in the 2D prior 
	conf_v1v2_pr = confidence(P_dMfbyMf_dchifbychif_pr_plot)
	s1_v1v2_pr = conf_v1v2_pr.height_from_level(0.5)
	s2_v1v2_pr = conf_v1v2_pr.height_from_level(0.9)

	# compute the credible level of the GR value from the combined posteriors
	conf_v1v2_joint = confidence(P_dMfbyMf_dchifbychif_joint)
	s1_v1v2_joint = conf_v1v2_joint.height_from_level(0.5)
	s2_v1v2_joint = conf_v1v2_joint.height_from_level(0.9)

	# Marginalization to one-dimensional joint_posteriors
	P_dMfbyMf_joint = np.sum(P_dMfbyMf_dchifbychif_joint, axis=0) * dy
	P_dchifbychif_joint = np.sum(P_dMfbyMf_dchifbychif_joint, axis=1) * dx

	# normalisation of joint posteriors
	P_dMfbyMf_joint /= np.sum(P_dMfbyMf_joint) * dx
	P_dchifbychif_joint /= np.sum(P_dchifbychif_joint) * dy

	# calculate the credible levels and intervals in the marginalized 1d posteriors (joint)
	s1_v1_joint, s2_v1_joint, left1_v1_joint, right1_v1_joint, left2_v1_joint, right2_v1_joint = calc_cred_intervals_in_1d(P_dMfbyMf_joint, dMfbyMf_vec)
	s1_v2_joint, s2_v2_joint, left1_v2_joint, right1_v2_joint, left2_v2_joint, right2_v2_joint = calc_cred_intervals_in_1d(P_dchifbychif_joint, dchifbychif_vec)

	# calculate the GR percentile from combined posterior  
	gr_height_joint = P_dMfbyMf_dchifbychif_joint[np.argmin(abs(dMfbyMf_vec)), np.argmin(abs(dchifbychif_vec))]
	gr_credib_level_joint = conf_v1v2_joint.level_from_height(gr_height_joint)
	print 'Credible level of the GR value (combined): %.1f%% '%(100.*gr_credib_level_joint)
	print '90%% credible intervals on dMfbyMfmean (combined): %2.3f, %2.3f '%(left2_v1_joint, right2_v1_joint)
	print '90%% credible intervals on dchifbychifmean (combined): %2.3f, %2.3f '%(left2_v2_joint, right2_v2_joint)

	# make plots 
	ax1.fill(dMfbyMf_vec, P_dMfbyMf_joint, 'gray', alpha=0.5)
	ax1.plot(dMfbyMf_vec, P_dMfbyMf_pr, 'k',  lw=0.5, ls=':')
	ax2.fill(P_dchifbychif_joint, dchifbychif_vec, 'gray', alpha=0.5)
	ax2.plot(P_dchifbychif_pr, dchifbychif_vec, 'k', lw=0.5, ls=':')
	ax3.contourf(dMfbyMf_vec,dchifbychif_vec,gf(P_dMfbyMf_dchifbychif_joint), levels=(s2_v1v2_joint, np.inf), linewidths=(1.5,0.5), colors='gray', linestyles=('--'), alpha=0.5)
	ax3.contour(dMfbyMf_vec,dchifbychif_vec,gf(P_dMfbyMf_dchifbychif_pr_plot), levels=(s2_v1v2_pr,), linewidths=(0.5,0.5), colors='k', linestyles=(':'))
	ax3.plot(0, 0, 'k+', mew=1)
	# axis limits 
	ax1.set_xlim(-1.5,1.5)
	ax1.set_ylim(0,6)
	ax2.set_ylim(-1,1)
	ax2.set_xlim(0,5)
	ax3.set_xlim(-1.5,1.5)
	ax3.set_ylim(-1,1)
	# tick labels 
	ax1.xaxis.tick_top()
	ax2.yaxis.tick_right()
	ax1.set_xticks(np.arange(-1.5, 1.51, 0.5))
	ax1.set_yticks(np.arange(1, 7, 1.))
	ax2.set_xticks(np.arange(1, 6, 1.))
	ax2.set_yticks(np.arange(-1, 1.1, 0.5))
	ax3.set_xticks(np.arange(-1.5, 1.51, 0.5))
	ax3.set_yticks(np.arange(-1, 1.1, 0.5))
	# axis labels and legends 
	ax1.legend(bbox_to_anchor=(1.05, 1.1), loc=2, borderaxespad=0.,frameon=False)
	ax2.set_xlabel(r'$P(\Delta a_{\mathrm{f}}/ \bar{a}_{\mathrm{f}})$', labelpad=8)
	ax1.set_ylabel(r'$P(\Delta M_{\mathrm{f}}/ \bar{M}_{\mathrm{f}})$', labelpad=10)
	ax3.set_xlabel(r'$\Delta M_{\mathrm{f}}/ \bar{M}_{\mathrm{f}}$', labelpad=8)
	ax3.set_ylabel(r'$\Delta a_{\mathrm{f}}/ \bar{a}_{\mathrm{f}}$', labelpad=-3)
	plt.subplots_adjust(wspace=0., hspace=0.)
	# save plots 
	plt.savefig('imrtgr_O1O2_%s.pdf'%approx)
